/*  Last edited: Dec  5 12:28 1995 (esr) */

/*
   MKDIV.C
   Erik Sonnhammer

  Date        Modification
--------  ----------------------------------------------------
93        Created
95-12-05  Added -f <db_file> for easy use on fasta files.
*/

#include "dbidx.h"

int WriteHeader(FILE *file, Header *head)
{
  int i;

  rewind(file);

  if (Reversebytes) reversebytes((char *)&head->file_size, 4);
  if (fwrite(&head->file_size, 4, 1, file) != 1) return NULL;
  if (Reversebytes) reversebytes((char *)&head->file_size, 4);

  if (Reversebytes) reversebytes((char *)&head->nrecords, 4);
  if (fwrite(&head->nrecords, 4, 1, file) != 1) return NULL;
  if (Reversebytes) reversebytes((char *)&head->nrecords, 4);

  if (Reversebytes) reversebytes((char *)&head->record_size, 2);
  if (fwrite(&head->record_size, 2, 1, file) != 1) return NULL;
  if (Reversebytes) reversebytes((char *)&head->record_size, 2);

  if (fwrite(head->db_name, 20, 1, file) != 1) return NULL;

  if (fwrite(head->db_relnum, 10, 1, file) != 1) return NULL;

  if (fwrite(head->db_reldate, 4, 1, file) != 1) return NULL;

  /* Fill up the empty 256 bytes */
  for (i=0; i<256; i++) fputc('\0', file);


  return 1;
}


int WriteDiv(FILE *file, Division *divrec)
{
  if (Reversebytes) reversebytes((char *)&divrec->div_code, 2);
  if (fwrite(&divrec->div_code, 2, 1, file) != 1) return NULL;

  if (fwrite(divrec->file_name, 12, 1, file) != 1) return NULL;

  return 1;
}


void main(int argc, char *argv[])
{
    FILE          *div;
    int            optc, year, month, day;
    unsigned short i;
    char           dbfile[21] = "";

    extern int   optind;
    extern char *optarg;
    char        *optstring="f:h"; 
    
    Header      *head;
    Division   **divrecs;
    Division     divrec;

    static char  usage[] = "\n\
  MKDIV - make division lookup file division.lkp for EMBL indices.\n\
\n\
  Usage: mkdiv [-options]\n\
\n\
    Options:\n\
    -h          Help\n\
    -f <file>   Don't prompt, but make a division.lkp for a single\n\
                fasta format file with dummy names and dates.\n\
\n\
\n";


#ifdef NOREVBYTES
  Reversebytes = 0;
#endif

    /* parse command line */
    while ((optc = getopt(argc, argv, optstring)) != -1)
	switch (optc) {
          
	case 'h':
	    printf("\n%s\n", usage); 
	    exit(1);                            break;
	case 'f':
	    memset(&dbfile, 0, 20);
	    strncpy(dbfile, optarg, 20);        break;
	}


    /* Write the records to the file */
    if ((div = fopen("division.lkp", "wr")) == NULL)
	fatal("Could not open the file division.lkp");
    
    if ((head = (Header *)malloc(sizeof(Header))) == NULL) fatal("malloc error");
    head->record_size = 14;

    if (*dbfile) {
	strcpy(head->db_name, dbfile);
	strcpy(head->db_relnum, "0");
	head->db_reldate->year  = 0;
	head->db_reldate->month = 0;
	head->db_reldate->day   = 0;
	head->nrecords = 1;
	head->file_size = 300 + head->nrecords*head->record_size;
	if (!WriteHeader(div, head)) 
	    fatal("Error writing Header");
	
	memset(&divrec, 0, sizeof(Division));
	divrec.div_code = 1;
	strcpy(divrec.file_name, dbfile);
	WriteDiv(div, &divrec);
    }
    else {
	printf("\nMKDIV - make a division lookup table.\n\n");

	printf("Database name (max 20) : "); scanf("%s", head->db_name);
	printf("Release nr             : "); scanf("%s", head->db_relnum);
	printf("Release date [yy mm dd]: "); scanf("%d%d%d", &year, &month, &day);
	head->db_reldate->year  = (unsigned short)year;
	head->db_reldate->month = (unsigned short)month;
	head->db_reldate->day   = (unsigned short)day;
	printf("Number of records      : "); scanf("%u", &head->nrecords);
	head->file_size = 300 + head->nrecords*head->record_size;
	if (!WriteHeader(div, head)) 
	    fatal("Error writing Header");
	
	/* Allocate all records */
	if (( divrecs = (Division **)malloc( head->nrecords*sizeof(Division *) )) == NULL) 
	    fatal("malloc error");
	
	for (i=0; i < head->nrecords; i++) {
	    if (( divrecs[i] = (Division *)malloc( sizeof(Division) )) == NULL) 
		fatal("malloc error");
	    
	    divrecs[i]->div_code = i+1;
	    printf("File name for division code %u  [max 12 chars] : ", i+1);
	    scanf("%s", divrecs[i]->file_name);
	}
	
	for (i=0; i < head->nrecords; i++) WriteDiv(div, divrecs[i]);
    }

    /* Clean up */
    fclose(div);
}
